@@ -1,15 +1,21 @@
 import os
+import re
 import sys
 import dill as pickle  # standard pickle doesn't handle lambdas
 import argparse
 import json
+import zlib
 import errno
 import shutil
+import docker
+import logging
+import binascii
 
 from uuid import uuid4
 from copy import deepcopy
 from jsonmerge import merge
 from urllib.parse import urlsplit
+from tempfile import NamedTemporaryFile
 from typing import MutableMapping, MutableSequence
 from ruamel.yaml.comments import CommentedMap
 from airflow.configuration import (
@@ -46,7 +52,10 @@ from cwl_airflow.utilities.helpers import (
     dump_json,
     get_absolute_path,
     get_rootname,
-    remove_field_from_dict
+    remove_field_from_dict,
+    get_uncompressed,
+    get_compressed,
+    get_files
 )
 
 
@@ -65,10 +74,77 @@ CWL_ENABLE_DEV = True      # better not to change without need
 
 DAG_TEMPLATE="""#!/usr/bin/env python3
 from cwl_airflow.extensions.cwldag import CWLDAG
-dag = CWLDAG(workflow="{0}", dag_id="{1}")
+dag = CWLDAG(
+    workflow="{0}",
+    dag_id="{1}"
+)
 """
 
 
+def overwrite_deprecated_dag(
+    dag_location,
+    deprecated_dags_folder=None
+):
+    """
+    Loads DAG content from "dag_location" file. Searches for "dag.create()" command.
+    If not found, we don't need to upgrade this DAG (it's either not from CWL-Airflow,
+    or already in a new format). If "deprecated_dags_folder" is not None, copies original
+    DAG file there before DAG upgrading. After copying deprecated DAG to the
+    "deprecated_dags_folder" updates ".airflowignore" with DAG file basename to exclude
+    it from Airflow parsing. Upgraded DAG will always include base64 encoded zlib
+    compressed workflow content. In case "workflow_location" is relative path, it will
+    be resolved based on the dirname of "dag_location" (useful for tests only, because
+    all our old DAGs always have absolute path to the CWL file). Function doesn't backup
+    or update the original CWL file.
+    TODO: in case more coplicated DAG files that include "default_args", etc, this function
+    should be updated to the more complex one.
+    """
+
+    with open(dag_location, "r+") as io_stream:                # open for both reading and writing
+
+        dag_content = io_stream.read()
+
+        if not re.search("dag\\.create\\(\\)", dag_content):   # do nothing if it wasn't old-style DAG
+            return
+
+        workflow_location = get_absolute_path(                 # resolve relative to dirname of "dag_location" (good for tests)
+            re.search(
+                "(cwl_workflow\\s*=\\s*[\"|'])(.+?)([\"|'])",
+                dag_content
+            ).group(2),
+            os.path.dirname(dag_location)
+        )
+
+        dag_id = re.search(
+            "(dag_id\\s*=\\s*[\"|'])(.+?)([\"|'])",
+            dag_content
+        ).group(2)
+
+        compressed_workflow_content = get_compressed(
+            fast_cwl_load(workflow_location)                   # no "run" embedding or convertion to Workflow. If DAG worked, cwl should be ok too
+        )
+
+        if deprecated_dags_folder is not None:                 # copy old DAG to the folder with deprecated DAGs, add ".airflowignore"
+            get_dir(deprecated_dags_folder)                    # try to create "deprecated_dags_folder" if it doesn't exist
+            shutil.copy(dag_location, deprecated_dags_folder)  # copy DAG file
+            ignore = os.path.join(
+                deprecated_dags_folder,
+                ".airflowignore"
+            )
+            with open(ignore, "a") as output_stream:           # add deprecated DAG to ".airflowignore"
+                output_stream.write(
+                    os.path.basename(dag_location) + "\n"
+                )
+
+        io_stream.seek(0)                                      # rewind "dag_location" file to the beginning
+        io_stream.write(
+            DAG_TEMPLATE.format(
+                compressed_workflow_content, dag_id
+            )
+        )
+        io_stream.truncate()                                   # remove old data at the end of a file if anything became shorter than original
+
+
 def conf_get(
     section,
     key,
@@ -249,6 +325,53 @@ def relocate_outputs(
     return relocated_job_data, workflow_report
 
 
+def get_containers(job_data, task_id):
+    """
+    Searches for cidfiles in the "step_tmp_folder", loads
+    container IDs from the found files, adds them to dict
+    in a form of {cid: location}. If nothing found,
+    returns {}.
+    """
+
+    containers = {}
+
+    step_tmp_folder, _, _, _ = get_temp_folders(
+        task_id=task_id,
+        job_data=job_data
+    )
+
+    for location in get_files(step_tmp_folder, ".*\\.cid$").values():
+        try:
+            with open(location, "r") as input_stream:
+                containers[input_stream.read()] = location
+        except OSError as err:
+            logging.error(f"Failed to read container ID \
+                from {location} due to \n{err}")
+
+    return containers
+
+
+def kill_containers(containers):
+    """
+    Iterates over "containers" dictionary received from "get_containers"
+    and tries to kill all running containers based on cid. If killed
+    container was not in "running" state, was successfully killed or not
+    found at all, removes correspondent cidfile.
+    """
+
+    docker_client = docker.from_env()
+    for cid, location in containers.items():
+        try:
+            container = docker_client.containers.get(cid)
+            if container.status == "running":
+                container.kill()
+            os.remove(location)
+        except docker.errors.NotFound as err:
+            os.remove(location)
+        except docker.errors.APIError as err:
+            logging.error(f"Failed to kill container. \n {err}")
+
+
 def execute_workflow_step(
     workflow,
     task_id,
@@ -720,7 +843,12 @@ def slow_cwl_load(workflow, cwl_args=None, only_tool=None):
     because the whole Workflow object later fails to be unpickled).
     If "workflow" was already parsed into CommentedMap, return it
     unchanged (in a form similar to what we can get if parsed
-    "workflow" with "only_tool" set to True)
+    "workflow" with "only_tool" set to True).
+    If "workflow" was a zlib compressed content of a file, it needs
+    to be uncompressed, then written to the temp file and loaded the
+    same way as described above. After loading temp file will be
+    removed automatically. First always try to uncompress, because
+    it's faster.
     """
     
     cwl_args = {} if cwl_args is None else cwl_args
@@ -731,14 +859,26 @@ def slow_cwl_load(workflow, cwl_args=None, only_tool=None):
 
     default_cwl_args = get_default_cwl_args(cwl_args)
 
-    workflow_data = load_tool(
-        workflow,
-        setup_loadingContext(
-            LoadingContext(default_cwl_args),
-            RuntimeContext(default_cwl_args),
-            argparse.Namespace(**default_cwl_args)
+    def __load(location):
+        return load_tool(
+            location,
+            setup_loadingContext(
+                LoadingContext(default_cwl_args),
+                RuntimeContext(default_cwl_args),
+                argparse.Namespace(**default_cwl_args)
+            )
         )
-    )
+
+    try:
+        with NamedTemporaryFile(mode="w") as temp_stream:  # guarantees that temp file will be removed
+            json.dump(
+                get_uncompressed(workflow, parse_as_yaml=True),
+                temp_stream
+            )
+            temp_stream.flush()                            # otherwise it might be only partially written
+            workflow_data = __load(temp_stream.name)
+    except (zlib.error, binascii.Error):                   # file was real
+        workflow_data = __load(workflow)
 
     return workflow_data.tool if only_tool else workflow_data
 
